#include "config.h"
#include "macros.h"
#include <sys/types.h>
#include <sys/stat.h>
#include "network.h"
#include "readconf.h"
#include "ag.h"
#include "share.h"
#include "redirect.h"
#include "filetransfer.h"
#include "ag_fnc.h"
#include "queue.h"


int srv_s=-1,ft_nld=0;
struct sockaddr_in ag_srv_addr;


/* closes the server connection */
void close_server_conn()
{
	if(srv_s != -1)
		close(srv_s);
	srv_s = -1;
	status = AG_STAT_DISCONNECTED;
	shared = 0;

	/* empty the send queue */
	queue_empty();

	/* close all filetransfers */
	ft_close_all();

	if(ag_loginstr)
		ag_free(ag_loginstr);

	/* trigger event */
	event_action("disconnect", 0);
}


unsigned char *get_ft_bar(int pos, int size)
{
	static unsigned char bar[11];
	int b;

	b = rdiv(pos*11, size)-1;

	memcpy(bar, "          ", 11);

	if(b >= 0 && b < 10)
		bar[b] = '>';
	while(b > 0) bar[--b] = '-';

	return bar;
}


void write_status_file()
{
	struct timeval tv;
	unsigned char *statstr;
	unsigned char tmpfile[MAX_PATH];
	FILE *fd;
	int i;

	if(!conf.statusfile[0])
		return;

	context;

	if(snprintf(tmpfile, MAX_PATH, "%s.tmp", conf.statusfile) == -1)
	{
		NOTICE("Statusfile path truncated!");
		return;
	}

	fd = fopen(tmpfile, "w");
	if(!fd)
	{
		NOTICE("Couldn't open '%s': %s", tmpfile, ERRMSG());
		return;
	}

	if(status == AG_STAT_DISCONNECTED)
		statstr = "DISCONNECTED";
	else if(status == AG_STAT_CONNECTING)
		statstr = "CONNECTING";
	else if(status == AG_STAT_CONNECTED)
		statstr = "CONNECTED";
	else if(status == AG_STAT_SEND_SHARES)
		statstr = "SEND_SHARES";
	else if(status == AG_STAT_LOGGED_IN)
		statstr = "LOGGED_IN";
	else
		statstr = "UNKNOWN";

	fprintf(fd, "status %s\n", statstr);
	fprintf(fd, "server %s %d\n", inet_ntoa(ag_srv_addr.sin_addr), htons(ag_srv_addr.sin_port));
	fprintf(fd, "share %d %d %d %d %d\n", old_shares, new_shares, upd_shares, del_shares, shared);
	fprintf(fd, "loginstr %s\n", ag_loginstr);
	gettimeofday(&tv, NULL);
	fprintf(fd, "uptime %d\n", (int)(tv.tv_sec - uptime.tv_sec));
	for(i=0;i<MAX_FT;i++)
	{
		if(ft[i].status != FT_ACTIVE)
			continue;

		if(ft[i].direction == AG_FT_RECV)
		{
			fprintf(fd, "ftrecv %s %s %d %d %d %d: %s/%s\n",
						ft[i].connecting == AG_FT_LISTEN ? "listen" : "connect",
						ft[i].ip, ft[i].port, (int)ftell(ft[i].fd), ft[i].filesize, ft[i].speed,
						conf.download_dir, ft[i].file);
		}
		else
		{
			fprintf(fd, "ftsend %s %s %d %d %d %d: %s\n",
						ft[i].connecting == AG_FT_LISTEN ? "listen" : "connect",
						ft[i].ip, ft[i].port, (int)ftell(ft[i].fd), ft[i].filesize, ft[i].speed,
						ft[i].ufile);
		}
	}

	fclose(fd);

	if(rename(tmpfile, conf.statusfile) == -1)
		NOTICE("Couldn't move '%s' to '%s': %s", tmpfile, conf.statusfile, ERRMSG());
}


/* fetches a server from the redirect server */
int ag_poll_redirect()
{
	static int s = -1;
	int n;

	if(s == -1)
	{
		context;
		/* fetch a server from the redirect server */
		status = AG_STAT_DISCONNECTED;
		NOTICE("Fetching server from " AG_REDIRECT_SERVER "...");
		/* resolve will block - usually not a problem */
		s = tcp_connect(make_sockaddr(AG_REDIRECT_SERVER, 21), 1);
		if(s == -1)
		{
			NOTICE("Couldn't connect to redirect server: %s", ERRMSG());
			return 0;
		}
	}

	context;

	n = poll_sock(s, 0);
	if(n == -1)
	{
		NOTICE("poll_sock() error: %s", ERRMSG());
		close(s);
		s = -1;
		return 0;
	}

	if(n == 0)
		return 1;

	context;

	if(ag_fetch_redirect(s, &ag_srv_addr))
		status = AG_STAT_CONNECTING;

	close(s);
	s = -1;

	return 1;
}




/*
	receives server packets and sends them off to ag_handle_packet
	(unless the packet is a filetransfer packet)
*/

int ag_poll_server()
{
	static struct sharestruct *share = NULL;
	static unsigned char rbuf[16384];
	static int rbufs=0;
	static struct timeval tv_ping,tv_conn,tv;
	int i,n=0,psize;


	if(status != AG_STAT_DISCONNECTED && status != AG_STAT_CONNECTING)
	{
		/* ping the server */
		context;
		gettimeofday(&tv, NULL);
		if(tv_ping.tv_sec+PINGFREQ < tv.tv_sec)
		{
			if(ag_send_status() == -1)
			{
				NOTICE("Error when pinging server: %s", ERRMSG());
				close_server_conn();
				return 0;
			}
			gettimeofday(&tv_ping, NULL);

			LOG_DISABLE();
			NOTICE("Ping! - Status: %d download%s, %d upload%s, %d shared file%s.",
					ft_downloads, ft_downloads==1?"":"s",
					ft_uploads, ft_uploads==1?"":"s",
					shared, shared==1?"":"s");

			for(i=0;i<MAX_FT;i++)
			{
				if(ft[i].status == FT_ACTIVE)
				{
					int pos = ft[i].fd ? ftell(ft[i].fd) : 0;
					int size = ft[i].filesize;

					if(pos > size)
						size = pos;

					NOTICE("%s: [%s] %.1f%% (%.2fKB of %.2fKB @ %.2fKB)",
						ft[i].direction == AG_FT_RECV ? "Download" : "  Upload",	/* Download / Upload */
						get_ft_bar(pos, size),										/* Progress bar */
						(float)pos / size * 100.0f,									/* Percent completed */
						(float)pos / 1024.0f,										/* Current position */
						(float)ft[i].filesize / 1024.0f,							/* Total filesize */
						(float)ft[i].speed / 1024.0f);								/* Speed */

					NOTICE("        > %s", ft[i].file);	/* filename */
				}
				else if(ft[i].status != FT_UNUSED)
				{
					NOTICE("%s: Setting up...",
						ft[i].direction == AG_FT_RECV ? "Download" : "  Upload");
				}
			}
			NOTICE("----");
			LOG_ENABLE();
		}
	}

	if(status == AG_STAT_CONNECTING && srv_s == -1)
	{
		/* connect to the server */
		context;
		NOTICE("Connecting to %s:%d...", inet_ntoa(ag_srv_addr.sin_addr), htons(ag_srv_addr.sin_port));

		gettimeofday(&tv_conn, NULL);
		srv_s = tcp_connect(&ag_srv_addr, 1);
		if(srv_s == -1)
		{
			NOTICE("Couldn't connect to server: %s", ERRMSG());
			close_server_conn();
			return 0;
		}
	}


	if(status == AG_STAT_CONNECTING)
	{
		/* request magic */
		context;
		DEBUG("request magic");
		if(ag_srv_msg(AG_MSG_MAGICREQUEST, NULL, 0, QPRIO_HIGH) == -1)
		{
			if(errno == EAGAIN || errno == ENOTCONN)
			{
				gettimeofday(&tv, NULL);
				if(tv_conn.tv_sec+CONNTIMEOUT < tv.tv_sec)
				{
					NOTICE("ZzZzZzzz, trying another server.");
					close_server_conn();
				}
				return 1;
			}
			else
			{
				close_server_conn();
				return 1;
			}
		}
		else
		{
			NOTICE("Connected. Logging in...");
			status = AG_STAT_CONNECTED;
			gettimeofday(&tv_ping, NULL);
			rbufs = 0;

			/* trigger event */
			event_action("connect", 2,
					inet_ntoa(ag_srv_addr.sin_addr),
					int2str(htons(ag_srv_addr.sin_port))
				);
		}
		return 1;
	}


	if(status == AG_STAT_SEND_SHARES)
	{
		struct sharestruct *next_share;
		int cost=0;

		/* send shares */
		if(share == NULL)
			share = sharelist;

		context;
		/* if there are any active filetransfers, lower the max cost */
		while(cost < ((ft_downloads+ft_uploads) ? 100 : 1000))
		{
			if(share) {
				next_share = share->next;

				switch(share->status)
				{
				case SHARE_STAT_NEW:
					ag_send_new_share(share, &cost);
					break;
				case SHARE_STAT_OLD:
					ag_send_old_share(share, &cost);
					shared++;
					break;
				case SHARE_STAT_UPD:
					ag_send_upd_share(share, &cost);
					break;
				case SHARE_STAT_DEL:
					/* This doesn't work, but it doesn't really matter. */
					/*ag_send_old_share(share, *cost);
					ag_send_del_share(share, *cost);*/
					share_del(share);
					save_share_db = 1;
					del_shares--;
					break;
				}
				share = next_share;
			}

			if(!share)
				break;
		}

		if(!share)
		{
			status = AG_STAT_LOGGED_IN;
			NOTICE("Shares sent to server.");
		}
	}


	context;
	/* poll server socket */
	n = poll_sock(srv_s, 0);
	if(n == -1)
	{
		NOTICE("select() error, shutting down server connection.");
		close_server_conn();
		return 0;
	}

	if(n == 0)
		return 1;


	context;
	/* read from server socket */
	errno = 0;
	n = recv(srv_s, &rbuf[rbufs], sizeof(rbuf)-rbufs, 0);
	if(n <= 0)
	{
		NOTICE("Error while receiving data from server: %s", ERRMSG());
		close_server_conn();
		return 0;
	}
	rbufs += n;

	/* parse packet buffer */
	while(rbufs > 4)
	{
		context;
		/* does the packet have a valid header? */
		if(!AG_VALID_HDR(rbuf))
		{
			NOTICE("Invalid packet header '%c%c'", rbuf[0], rbuf[1]);
			close_server_conn();
			return 0;
		}

		/* do we have a whole packet? */
		if(AG_WORD(&rbuf[2]) <= rbufs-2)
		{
			context;
			psize = AG_WORD(&rbuf[2]);	/* packet size */

			if(rbuf[4] != AG_MSG_FILETRANSFER)
			{
				context;
				/* handles all packets except for filetransfer packets */
				ag_handle_packet(psize-3, rbuf[4], &rbuf[5]);
			}
			else
			{
				context;
				/* handle filetransfer packets */
				i = ft_alloc();
				if(i == -1)
				{
					NOTICE("No free filetransfer slots, aborting transfer!");
					/* hack hack hack */
					if(AG_WORD(&rbuf[5]) < psize-3)
						ft_send_state(AG_FT_STATE_ERROR, &rbuf[7], AG_WORD(&rbuf[5]));
				}
				else
				{
					/* fill the allocated ft_struct with crap */
					if(ft_fillstruct(i, psize, &rbuf[5]) == -1)
					{
						ft_free(i);
						goto fterr;
					}

					context;
					if(ft[i].connecting == AG_FT_CONNECT)
					{
						/* connecting */
						NOTICE("Connecting to %s:%d...", ft[i].ip, ft[i].port);
						ft[i].s = tcp_connect(make_sockaddr(ft[i].ip, ft[i].port), 1);
						if(ft[i].s == -1)
						{
							/* socket error - notify the server and free ft_struct */
							ft_send_state(AG_FT_STATE_ERROR, ft[i].fileid, ft[i].fileid_len);
							ft_free(i);
						}
						else
						{
							ft_new_fd(ft[i].s);
							ft[i].fdset = AG_FD_WRITE;
							ft[i].status = FT_CONNECTING;
						}
					}
					else
					{
						/* listening */
						ft[i].s = tcp_listen(make_sockaddr("0.0.0.0", ft[i].port), 4);
						if(ft[i].s == -1)
						{
							/* listen error - notify server */
							ft_send_state(AG_FT_STATE_ERRBIND, ft[i].fileid, ft[i].fileid_len);
							ft_free(i);
						}
						else
						{
							ft_new_fd(ft[i].s);
							NOTICE("Listening on port %d...", ft[i].port);
							ft[i].fdset = AG_FD_READ;
							ft[i].status = FT_LISTENING;
						}
					}	/* if(ft[i].connecting) */

					/* no errors */
					if(ft[i].status != FT_UNUSED)
					{
						DEBUG("AG_FT_STATE_SETUP");
						if(ft[i].direction == AG_FT_RECV)
							ft_downloads++;
						else
							ft_uploads++;
						ft_send_state(AG_FT_STATE_SETUP, ft[i].fileid, ft[i].fileid_len);
						gettimeofday(&ft[i].timeout, NULL);
						if(ft[i].fdset == AG_FD_READ)
							FD_SET(ft[i].s, &ft_read_fd);
						else
							FD_SET(ft[i].s, &ft_write_fd);
					}
fterr:;
				}

			}	/* if(rbuf[4] != AG_MSG_FILETRANSFER) */

			/* re-arrange buffer */
			memcpy(rbuf, rbuf+psize+2, rbufs-(psize+2));
			rbufs -= psize+2;
		}
		else
		{
			/* we don't have a whole packet, break the loop */
			break;
		}
	}	/* if(rbufs > 4) */

	return 1;
}



int bandwidth_poll()
{
	static struct timeval bw_tv, tv;
	static int first=1;
	static int bw_down, bw_up, bw_diff=0;
	int bw_down_share=0, bw_up_share=0;		/* how much each transfer should be given ~ average */
	int bw_down_use=0, bw_up_use=0;			/* how much bandwidth we used in total */
	int bw_down_nfast=0, bw_up_nfast=0;		/* # of transfers going faster than the average */
	int bw_down_dec=0, bw_up_dec=0;
	int bw_share=0, bw_div;					/* temp vars */
	int i;

	if(!( (conf.bandwidth_download && ft_downloads) ||
			(conf.bandwidth_upload && ft_uploads) ))
		return 0;

	context;

	if(first)
	{
		/* first run */
		first = 0;
		gettimeofday(&bw_tv, NULL);
	}

	/* see if ~1 second has elapsed since the last call */
	gettimeofday(&tv, NULL);
	bw_diff = DIFF_MSEC(bw_tv, tv);
	DEBUG("bw_diff %d", bw_diff);
	if(bw_diff < 1000)
		return 0;
	if(bw_diff <= 1250)
		bw_diff = 1250;	/* don't drift too much */

	gettimeofday(&bw_tv, NULL);
	bw_tv.tv_usec += (bw_diff - 1000) * 1000;
	if(bw_tv.tv_usec > 1000000)
	{
		bw_tv.tv_sec++;
		bw_tv.tv_usec -= 1000000;
	}

	/* calculate how much bandwidth we can use */
	if(ft_downloads)
	{
		bw_down = conf.bandwidth_download; /* * bw_diff) / 1000;*/
		bw_down_share = bw_down / ft_downloads;
	}
	if(ft_uploads)
	{
		bw_up = conf.bandwidth_upload; /* * bw_diff) / 1000;*/
		bw_up_share = bw_up / ft_uploads;
	}


	for(i=0;i<MAX_FT;i++)
	{
		if(ft[i].status != FT_ACTIVE)
			continue;	/* not active */

		if(ft[i].direction == AG_FT_RECV && !conf.bandwidth_download)
			continue;	/* download && no download limit */

		if(ft[i].direction == AG_FT_SEND && !conf.bandwidth_upload)
			continue;	/* upload && no upload limit */


		if(ft[i].direction == AG_FT_RECV)
		{
			bw_share = bw_down_share;
			bw_div = ft_downloads;
		}
		else
		{
			bw_share = bw_up_share;
			bw_div = ft_uploads;
		}


		if(ft[i].bw == 0 && ft[i].bw_last == 0)
		{
			/* first time it's getting bandwidth (ie. this transfer just started) */
			ft[i].bw = bw_share;
		}
		else if(ft[i].bw == 0)
		{
			/* this connection used up all it's bandwidth - try to give it more */
			ft[i].bw = ft[i].bw_last + bw_share / 20;
		}
		else
		{
			/* this connection didn't use up all it's bandwidth - lower it's speed */
			ft[i].bw = ft[i].bw_last - bw_share / 20;
			if(ft[i].bw < 32)
				ft[i].bw = 32;
		}


		/* mark it as fast if its using more than the average speed */
		if(ft[i].bw > bw_share)
		{
			if(ft[i].direction == AG_FT_RECV)
				bw_down_nfast++;
			else
				bw_up_nfast++;
		}

		/* add to total bw usage */
		if(ft[i].direction == AG_FT_RECV)
			bw_down_use += ft[i].bw;
		else
			bw_up_use += ft[i].bw;
	}


	/* are we using too much downstream or upstream? */
	if(bw_down_use > bw_down && bw_down_nfast)
		bw_down_dec = (bw_down_use - bw_down) / bw_down_nfast;
	if(bw_up_use > bw_up && bw_up_nfast)
		bw_up_dec = (bw_up_use - bw_up) / bw_up_nfast;


	for(i=0;i<MAX_FT;i++)
	{
		if(ft[i].status != FT_ACTIVE)
			continue;

		if(ft[i].direction == AG_FT_RECV && conf.bandwidth_download)
		{
			if(ft[i].bw > bw_share)
				ft[i].bw -= bw_down_dec;
			if(ft[i].bw < 32) ft[i].bw = 32;
			ft[i].bw_last = ft[i].bw;

			DEBUG("ft %d got %d bw", i, ft[i].bw);
			if(!FD_ISSET(ft[i].s, &ft_read_fd))
				FD_SET(ft[i].s, &ft_read_fd);
		}
		else if(ft[i].direction == AG_FT_SEND && conf.bandwidth_upload)
		{
			if(ft[i].bw > bw_share)
				ft[i].bw -= bw_up_dec;
			if(ft[i].bw < 32) ft[i].bw = 32;
			ft[i].bw_last = ft[i].bw;

			DEBUG("ft %d got %d bw", i, ft[i].bw);
			if(!FD_ISSET(ft[i].s, &ft_write_fd))
				FD_SET(ft[i].s, &ft_write_fd);
		}
	}

	return 1;
}



int ag_poll_filetransfers()
{
	unsigned char path[MAX_PATH];
	struct sockaddr_in accept_addr;
	struct timeval tv;
	struct sharestruct *share;
	struct stat st;
	int i,r,n,plen,tmp_s,nfd;

	context;

	/* get the highest filetransfer descriptor */
	nfd = ft_max_fd();
	if(nfd == 0)
	{
		/* no active filetransfers.. */
		sleep(1);
		return 1;
	}

	bandwidth_poll();

	tv.tv_sec = 0;
	tv.tv_usec = 200000;
	/*DEBUG("poll ft max_fd %d", nfd);*/
	n = select(nfd+1, &ft_read_fd, &ft_write_fd, NULL, &tv);
	if(n == -1)
	{
		if(errno != EINTR)
			NOTICE("Fatal select() error: %s", ERRMSG());
		NOTICE("Shutting down all filetransfers!");
		ft_close_all();
		close_server_conn();
		return 0;
	}
	/*DEBUG("poll ft - %d sets", n);*/


	gettimeofday(&tv, NULL);
	for(i=0;i<MAX_FT;i++)
	{
		if(ft[i].status == FT_UNUSED)
			continue;

		/* look for stalled transfers / connect timeouts */
		if(ft[i].timeout.tv_sec+TRANSTIMEOUT < tv.tv_sec)
		{
			NOTICE("Aborting transfer of '%s': %s", ft[i].file, 
					ft[i].status==FT_CONNECTING?"Connection timed out":"Transfer stalled");
			ft_send_state(AG_FT_STATE_NOCONN, ft[i].fileid, ft[i].fileid_len);
			ft_free(i);
			continue;
		}

		/* update transferspeed ... */
		if(ft[i].status == FT_ACTIVE)
			ft_update_speed(i);
	}

	/* zero sets returned */
	if(n == 0)
	{
		/* add all sockets back to their fdset */
		for(i=0;i<MAX_FT;i++)
		{
			if(ft[i].status == FT_UNUSED)
				continue;

			/* dont add it back if it has no bandwidth */
			if(ft[i].direction == AG_FT_RECV && ft[i].status == FT_ACTIVE &&
				ft[i].bw == 0 && conf.bandwidth_download)
				continue;
			if(ft[i].direction == AG_FT_SEND && ft[i].status == FT_ACTIVE &&
				ft[i].bw == 0 && conf.bandwidth_upload)
				continue;

			if(ft[i].fdset == AG_FD_READ)
				FD_SET(ft[i].s, &ft_read_fd);
			else
				FD_SET(ft[i].s, &ft_write_fd);
		}
		return 1;
	}


	for(i=0;i<MAX_FT;i++)
	{
		if(ft[i].status == FT_UNUSED)
			continue;

		if(ft[i].fdset == AG_FD_READ && !FD_ISSET(ft[i].s, &ft_read_fd))
		{
			/* add the socket back to the read fdset */
			if((ft[i].bw > 0 || !conf.bandwidth_download) || ft[i].status != FT_ACTIVE)
				FD_SET(ft[i].s, &ft_read_fd);
			continue;
		}
		if(ft[i].fdset == AG_FD_WRITE && !FD_ISSET(ft[i].s, &ft_write_fd))
		{
			/* add the socket back to the write fdset */
			if((ft[i].bw > 0 || !conf.bandwidth_upload) || ft[i].status != FT_ACTIVE)
				FD_SET(ft[i].s, &ft_write_fd);
			continue;
		}

		/* reset timeout */
		gettimeofday(&ft[i].timeout, NULL);

		switch(ft[i].status)
		{
		/* ======== LISTENING TRANSFERS ======== */
		case FT_LISTENING:
			/* someone connected to us */
			context;
			DEBUG("FT_LISTENING");

			/* accept connection */
			n = sizeof(struct sockaddr_in);
			tmp_s = accept(ft[i].s, (struct sockaddr *)&accept_addr, &n);
			if(tmp_s == -1)
			{
				ft_close_conn(i);
				break;
			}

			if(!ip_match(ft[i].ip, &accept_addr.sin_addr))
			{
				NOTICE("Warning: connection from IP '%s' does not match IP mask '%s', dropping...", inet_ntoa(accept_addr.sin_addr), ft[i].ip);
				close(tmp_s);
				continue;
			}
			NOTICE("Accepted connection from %s...", inet_ntoa(accept_addr.sin_addr));

			/* connection accepted - close listening socket */
			FD_CLR(ft[i].s, &ft_read_fd);
			ft_del_fd(ft[i].s);
			close(ft[i].s);

			socket_set_async(tmp_s, 1);
			ft[i].s = tmp_s;
			ft_new_fd(ft[i].s);
			ft[i].fdset = AG_FD_WRITE;
			FD_SET(ft[i].s, &ft_write_fd);

			ft[i].status = FT_CONNECTING;
			/* fall through */

		/* ======== CONNECTING TRANSFERS ======== */
		case FT_CONNECTING:
			/* connection established */
			context;
			DEBUG("FT_CONNECTING");

			if(ft[i].direction == AG_FT_RECV)
			{
				ft[i].fd = ft_fopen(ft[i].file, 1);
				if(!ft[i].fd)
				{
					ft_close_conn(i);
					break;
				}

				/* we are receiving - send request */
				ft[i].resume_pos = ftell(ft[i].fd);
				if(ft_send_request(i, ft[i].resume_pos) == -1)
				{
					ft_close_conn(i);
					break;
				}
			}
			ft[i].status = FT_WAIT;
			ft[i].fdset = AG_FD_READ;
			FD_CLR(ft[i].s, &ft_write_fd);
			FD_SET(ft[i].s, &ft_read_fd);
			break;


		/* ======== WAITING / LAST HANDSHAKE TRANSFERS ======== */
		case FT_WAIT:
			/* waiting for ack / request */
			context;
			DEBUG("FT_WAIT");

			if(!ft[i].buf)
			{
				/* allocate some memory for a receive buffer */
				ft[i].buf = ag_malloc(TRANSBUF);

				if(!ft[i].buf)
				{
					ft_close_conn(i);
					break;
				}
			}

			n = recv(ft[i].s, ft[i].buf, TRANSBUF, 0);
			if(n == -1)
			{
				NOTICE("Aborting transfer of '%s': %s", ft[i].file, ERRMSG());
				if(errno == ECONNRESET)
					ft_send_state(AG_FT_STATE_CLOSED, ft[i].fileid, ft[i].fileid_len);
				else
					ft_send_state(AG_FT_STATE_NOCONN, ft[i].fileid, ft[i].fileid_len);
				ft_free(i);
				break;
			}

			/* handshake */
			if(ft[i].direction == AG_FT_RECV)
			{
				/* got ACK */
				plen = ft_recv_response(i, n, ft[i].buf);
				if(plen == -1)
				{
					ft_close_conn(i);
					break;
				}

				if(plen > 0 && (n-plen) > 0)
				{
					/* save the overflow (we got more data than just the ACK packet...) */
					if(fwrite(ft[i].buf, n-plen, 1, ft[i].fd) == 0)
					{
						NOTICE("Error writing to disk: %s", ERRMSG());
						ft_close_conn(i);
						break;
					}
				}
			}
			else
			{
				/* got REQEUST */
				context;

				ft[i].fd = ft_fopen(ft[i].ufile, 0);
				if(!ft[i].fd)
				{
					NOTICE("Error opening '%s': %s", ft[i].ufile, ERRMSG());
					ft_close_conn(i);
					break;
				}

				/* receive request packet */
				plen = ft_recv_request(i, n, ft[i].buf);
				if(plen == -1)
				{
					ft_close_conn(i);
					break;
				}

				if(plen > 0)
				{
					/* respond */
					if(ft_send_response(i) == -1)
					{
						ft_close_conn(i);
						break;
					}

					/* move socket to ft_write_fd - we are sending */
					ft[i].fdset = AG_FD_WRITE;
					FD_CLR(ft[i].s, &ft_read_fd);
					FD_SET(ft[i].s, &ft_write_fd);
				}
			}
			ft[i].status = FT_ACTIVE;

			ft[i].bw = 0;		/* no initial bandwidth */
			ft[i].bw_last = 0;

			/* set timers */
			gettimeofday(&ft[i].start, NULL);
			gettimeofday(&ft[i].lastspd, NULL);

			NOTICE("Started %s of '%s'...", ft[i].direction == AG_FT_RECV ? "download" : "upload", ft[i].file);

			if(ft[i].direction == AG_FT_RECV)
				event_action("download_start", 3, conf.download_dir, ft[i].file, ft[i].ip);
			else
				event_action("upload_start", 3, ft[i].ufile, ft[i].file, ft[i].ip);

			break;


		/* ======== ACTIVE TRANSFERS ======== */
		case FT_ACTIVE:
			context;

			errno = 0;
			if(ft[i].direction == AG_FT_RECV)
			{
				/* receive data */
				context;

				if(ft[i].bw <= 0 && conf.bandwidth_download)
				{
					FD_CLR(ft[i].s, &ft_read_fd);
					continue;
				}

				if(ft[i].bw >= TRANSBUF || conf.bandwidth_download == 0)
					n = recv(ft[i].s, ft[i].buf, TRANSBUF, 0);
				else
					n = recv(ft[i].s, ft[i].buf, ft[i].bw, 0);

				if(n <= 0)
				{
					/* error... */
					if(ftell(ft[i].fd) < ft[i].filesize)
					{
						/* we didn't get the whole file */
						NOTICE("Download of '%s' aborted: %s", ft[i].file, ERRMSG());
						ft_send_state(AG_FT_STATE_CLOSED, ft[i].fileid, ft[i].fileid_len);
					}
					else
					{
						/* got the whole file! transfer complete. */
						ft_rename(i);
						NOTICE("Download of '%s' completed. Average speed was %.2fKB/sec.", ft[i].file, ft_avg_speed(i));
						ft_send_state(AG_FT_STATE_COMPLETE, ft[i].fileid, ft[i].fileid_len);

						/* share the downloaded file */
						if(conf.share_downloads)
						{
							share = share_new();
							if(!share)
							{
								NOTICE("Error allocating new share: %s", ERRMSG());
								goto share_err;
							}

							if(snprintf(path, MAX_PATH, "%s/%s", conf.download_dir, ft[i].file) == 1)
							{
								NOTICE("Path too long!");
								share_del(share);
								goto share_err;
							}

							if(stat(path, &st))
							{
								NOTICE("Couldn't stat '%s': %s", path, ERRMSG());
								share_del(share);
								goto share_err;
							}

							NOTICE("Sharing new downloaded file...");
							share->path = ag_strdup(conf.download_dir);
							share->file = ag_strdup(ft[i].file);
							share->hash = share_make_hash(conf.download_dir, ft[i].file);

							/*
								free the filetransfer slot here cause if we don't it will
								cause problems in ag_send_new_share() later
							*/
							ft_free(i);

							share->size = st.st_size;
							share->lastmod = st.st_mtime;
							share->status = SHARE_STAT_NEW;
							new_shares++;

							share->is_download = 1;
							share->download_speed = ft[i].speed;

							ag_send_new_share(share, /*whatever*/&r);
share_err:;
						}	/* if(conf.share_downloads) */
						else
						{
							/* trigger download_end event */
							event_action("download_end", 4, conf.download_dir, ft[i].file, float2str(ft_avg_speed(i)), "0");
						}
					}
					ft_free(i);
				}
				else
				{
					/* save what we received. :) */
					if(fwrite(ft[i].buf, n, 1, ft[i].fd) == 0)
					{
						NOTICE("Error while writing '%s': %s", ft[i].file, ERRMSG());
 						ft_close_conn(i);
					}
					else
					{
						ft_update_speed(i);
						if(conf.bandwidth_download)
							ft[i].bw -= n;
					}
				}
			}		/* ======== UPLOADS ======= */
			else	/* ======== UPLOADS ======= */
			{		/* ======== UPLOADS ======= */
				/* read data from disk */
				context;

				if(ft[i].bw <= 0 && conf.bandwidth_upload)
				{
					FD_CLR(ft[i].s, &ft_write_fd);
					continue;
				}

				if(ft[i].bw >= TRANSBUF || conf.bandwidth_upload == 0)
					r = fread(ft[i].buf, 1, TRANSBUF, ft[i].fd);
				else
					r = fread(ft[i].buf, 1, ft[i].bw, ft[i].fd);

				if(r > 0)
				{
					/* send it */
					n = send(ft[i].s, ft[i].buf, r, 0);
					if(n == -1)
					{
						if(errno == EAGAIN || errno == EWOULDBLOCK || errno == ENOBUFS)
						{
							/* can't send more - but select() said we could? */
							fseek(ft[i].fd, -r, SEEK_CUR);
							ft_update_speed(i);
						}
						else
						{
							/* send() error */
							NOTICE("Upload of '%s' aborted: %s", ft[i].file, ERRMSG());
							ft_close_conn(i);
						}
					}
					else
					{
						/* if all of the data was not sent, roll back */
						if(n > 0 && n < r)
						{
							fseek(ft[i].fd, -(r-n), SEEK_CUR);
							if(conf.bandwidth_upload)
								ft[i].bw -= (r-n);
						}
						else
						{
							if(conf.bandwidth_upload)
								ft[i].bw -= r;
						}
						/* no more allocated bandwidth - remove from fd_set */
						ft_update_speed(i);
					}
				}
				else if(!feof(ft[i].fd))
				{
					/* fatal read error */
					NOTICE("Error while reading '%s': %s", ft[i].file, ERRMSG());
					ft_close_conn(i);
				}
				else
				{
					/* end of file reached - filetransfer completed */
					NOTICE("Upload of '%s' completed. Average speed was %.2fKB/sec.", ft[i].file, ft_avg_speed(i));
					ft_send_state(AG_FT_STATE_COMPLETE, ft[i].fileid, ft[i].fileid_len);
					event_action("upload_end", 3, ft[i].ufile, ft[i].file, float2str(ft_avg_speed(i)));
					ft_free(i);
				}
			}	/* if(ft[i].direction == AG_FT_RECV) - else */
			break;
		}
		context;
	}

	return 1;
}





void ag_main_loop()
{
	struct timeval now,last,last_ws;

	context;

	/* init filetransfer structs */
	ft_init();

	gettimeofday(&last, NULL);
	gettimeofday(&last_ws, NULL);
	while(running)
	{
		if(status == AG_STAT_DISCONNECTED)
		{
			context;
			ag_poll_redirect();
		}

		if(!running)
			break;

		gettimeofday(&now, NULL);
		if(conf.statusfile[0] && DIFF_MSEC(last_ws, now) > conf.statusfile_interval)
		{
			write_status_file();
			gettimeofday(&last_ws, NULL);
		}

		/* only poll the server once every second */
		if(DIFF_MSEC(last, now) > 1000 && status != AG_STAT_DISCONNECTED)
		{
			ag_poll_server();
			share_interval_save_db();
			if(!queue_poll())
				close_server_conn();
			gettimeofday(&last, NULL);
		}

		if(!running)
			break;

		context;
		ag_poll_filetransfers();
	}


	if(status && srv_s != -1)
	{
		NOTICE("Logging out...");
		ag_srv_msg(AG_MSG_LOGOUT, NULL, 0, QPRIO_HIGH);
		close_server_conn();
	}

	ft_close_all();
}

