#include "config.h"
#include "macros.h"
#include <stdarg.h>

struct eventstruct *event_first = NULL;
struct eventstruct *event_last = NULL;


struct eventstruct *event_register(unsigned char *event, unsigned char *action)
{
	if(event_last == NULL)
	{
		event_first = event_last = ag_malloc(sizeof(struct eventstruct));
		if(!event_first) return NULL;
		event_first->next = NULL;
	}
	else
	{
		event_last->next = ag_malloc(sizeof(struct eventstruct));
		if(!event_last->next) return NULL;
		event_last = event_last->next;
		event_last->next = NULL;
	}
	event_last->event = ag_strdup(event);
	event_last->action = ag_strdup(action);

	return event_last;
}


void event_unregister_all()
{
	struct eventstruct *e;

	while(event_first)
	{
		e = event_first->next;
		if(event_first->event)
			ag_free(event_first->event);
		if(event_first->action)
			ag_free(event_first->action);
		ag_free(event_first);
		event_first = e;
	}

	return;
}


void event_action(unsigned char *event, int argc, /* strings only! */...)
{
	va_list	va;
	int i;
	char **argv;
	struct eventstruct *e;

	argv = ag_malloc(sizeof(char *) * (argc+2));
	if(!argv) return;

	va_start(va, argc);
	for(i=1;i<=argc;i++)
		argv[i] = va_arg(va, char *);
	argv[i] = NULL;
	va_end(va);

	e = event_first;
	while(e)
	{
		if(!strcmp(e->event, event))
		{
			argv[0] = e->action;
			switch(fork())
			{
			case -1:
				NOTICE("Unable to fork(): %s", ERRMSG());
				return;

			case 0:
				execvp(e->action, argv);
				exit(0);
			}
		}
		e = e->next;
	}

	ag_free(argv);

	return;
}

