#include "config.h"
#include "macros.h"
#include "network.h"
#include "ag.h"
#include "share.h"
#include "ag_fnc.h"
#include "filetransfer.h"
#include "queue.h"
#include "ogg_info.h"

struct struct_ft ft[MAX_FT];
int ft_uploads,ft_downloads, ft_highest_fd=0;
fd_set ft_read_fd,ft_write_fd;


void ft_init()
{
	int i;

	context;

	ft_uploads = 0;
	ft_downloads = 0;
	for(i=0;i<MAX_FT;i++)
	{
		ft[i].status = FT_UNUSED;
	}

	FD_ZERO(&ft_read_fd);
	FD_ZERO(&ft_write_fd);
}


int ft_alloc()
{
	int i;

	context;

	for(i=0;i<MAX_FT;i++)
	{
		if(ft[i].status == FT_UNUSED)
		{
			memset(&ft[i], 0, sizeof(struct struct_ft));
			ft[i].s = -1;
			return i;
		}
	}

	return -1;
}


void ft_free(int id)
{
	context;

	if(ft[id].status != FT_UNUSED)
	{
		if(ft[id].direction == AG_FT_RECV)
			ft_downloads--;
		else
			ft_uploads--;
		ft[id].status = FT_UNUSED;
		if(ft[id].s != -1)
		{
			ft_del_fd(ft[id].s);
			if(FD_ISSET(ft[id].s, &ft_read_fd))
				FD_CLR(ft[id].s, &ft_read_fd);
			if(FD_ISSET(ft[id].s, &ft_write_fd))
				FD_CLR(ft[id].s, &ft_write_fd);
			close(ft[id].s);
		}
		if(ft[id].fd)
		{
			if(ft[id].file && !strcmp(ft[id].file, "new proxy test"))
			{
				unsigned char file[MAX_PATH];
				snprintf(file, MAX_PATH-6, "%s/new proxy test", conf.download_dir);
				unlink(file);
				strcat(file, ".fags");
				unlink(file);
				errno = 0;
			}
			fclose(ft[id].fd);
		}
		if(ft[id].ip)
			ag_free(ft[id].ip);
		if(ft[id].fileid)
			ag_free(ft[id].fileid);
		if(ft[id].file)
			ag_free(ft[id].file);
		if(ft[id].ufile)
			ag_free(ft[id].ufile);
		if(ft[id].buf)
			ag_free(ft[id].buf);
	}
}


/* closes a filetransfer slot and sends AG_FT_STATE_ERROR to the server */
void ft_close_conn(int id)
{
	context;

	ft_send_state(AG_FT_STATE_ERROR, ft[id].fileid, ft[id].fileid_len);
	ft_free(id);
}


void ft_close_all()
{
	int i;

	for(i=0;i<MAX_FT;i++)
		ft_free(i);
}


void ft_update_speed(int id)
{
	struct timeval tv;
	int timediff, recv;

	gettimeofday(&tv, NULL);

	timediff = DIFF_MSEC(ft[id].lastspd, tv);

	if(timediff >= 4000)
	{
		recv = ftell(ft[id].fd) - ft[id].lastspd_pos;
		ft[id].speed = recv * 1000 / timediff;

		ft[id].lastspd_pos = ftell(ft[id].fd);
		gettimeofday(&ft[id].lastspd, NULL);
	}

	return;
}


int ft_max_fd()
{
	return ft_highest_fd;
}


void ft_new_fd(int fd)
{
	if(fd > ft_highest_fd)
		ft_highest_fd = fd;

	DEBUG("high fd %d", ft_highest_fd);
}


void ft_del_fd(int fd)
{
	int i;

	DEBUG("delete fd %d", fd);

	if(fd == ft_highest_fd)
	{
		ft_highest_fd = 0;
		for(i=0;i<MAX_FT;i++)
		{
			if(ft[i].status == FT_UNUSED)
				continue;
			if(ft[i].s > ft_highest_fd && ft[i].s != fd)
				ft_highest_fd = ft[i].s;
		}
		DEBUG("new high fd %d", ft_highest_fd);
	}
}


int ft_find_by_fileid(unsigned char *fileid, int fileid_len)
{
	int i;

	context;

	for(i=0;i<MAX_FT;i++)
	{
		if(ft[i].status != FT_UNUSED && ft[i].fileid_len == fileid_len)
		{
			if(!memcmp(ft[i].fileid, fileid, fileid_len))
				return i;
		}
	}

	return -1;
}


/* returns the average speed of at transfer in KB/s, should be called when the transfer has been completed */
float ft_avg_speed(int id)
{
	struct timeval end;
	float total_time, avg_speed;

	gettimeofday(&end, NULL);

	/* calculate average speed */
	end.tv_sec -= ft[id].start.tv_sec;
	end.tv_usec -= ft[id].start.tv_usec;
	total_time = (float)end.tv_sec + (float)end.tv_usec/1000000.0f;
	avg_speed = (float)(ft[id].filesize-ft[id].resume_pos) / total_time / 1024.0f;

	return avg_speed;
}


/* "blah - blah blah.mp3.fags" -> "blah - blah blah.mp3" */
void ft_rename(int id)
{
	unsigned char *src,*dst;
	int l,c=1;
#ifdef USE_VORBIS
	struct sharestruct dummy_share;
#endif

	if(!*ft[id].file)
		return;

	l = strlen(conf.download_dir)+1+strlen(ft[id].file);

	src = ag_malloc(l + /* '.fags\0'  */ 6);
	if(!src) return;
	sprintf(src, "%s/%s.fags", conf.download_dir, ft[id].file);

	if(conf.capitalize_files)
		capitalize_filename(ft[id].file);

	dst = ag_malloc(l + /* '.999\0' */ 5);
	if(!dst)
	{
		ag_free(src);
		return;
	}
	sprintf(dst, "%s/%s", conf.download_dir, ft[id].file);

#ifdef USE_VORBIS
	/* see if the downloaded file is an ogg file  */
	if(l > 3 && !strcasecmp(&dst[l-3], "ogg"))
	{
		if(ogg_fill_info(&dummy_share, src, 1))
		{
			dst[l-3] = 'o';
			dst[l-2] = 'g';
			dst[l-1] = 'g';
		}
	}
#endif

	if(!file_exists(dst))
	{
		rename(src, dst);
		ag_free(src);
		ag_free(dst);
		return;
	}

	while(c < 1000)
	{
		sprintf(dst, "%s/%s.%d", conf.download_dir, ft[id].file, c);
		if(!file_exists(dst)) {
			rename(src, dst);
			ag_free(src);
			ag_free(dst);
			return;
		}
		c++;
	}
	NOTICE("Failed to rename '%s'.", ft[id].file);

	ag_free(src);
	ag_free(dst);
}


/* opens a file & removes bad unsigned chars from the filname */
FILE *ft_fopen(unsigned char *file, int rw)
{
	unsigned char *str;
	FILE *fd;
	int len;

	context;

	if(rw)
	{
		if(!*file)
		{
			return NULL;
		}
	
		str = file;
		while(*str)
		{
			/* filter out stupid chars like  '?', '*', '$' and '~' */
			if(*str < 32 || *str == '/' ||
				*str == '?' || *str == '*' ||
				*str == '~' || *str == '$')
				*str = '_';
			str++;
		}
		/* also, files starting with a dash (-) can be quite annoying ;) */
		if(*file == '-')
			*file = '_';

		len = strlen(conf.download_dir)+1 + strlen(file)+6;
		str = ag_malloc(len);
		if(!str) return NULL;
		snprintf(str, len, "%s/%s.fags", conf.download_dir, file);

		fd = fopen(str, "a+");
		if(!fd)
			NOTICE("Unable to open '%s': %s", str, ERRMSG());
		ag_free(str);
	}
	else
	{
		fd = fopen(file, "r");
	}

	return fd;
}



int ft_fillstruct(int id, int len, unsigned char *data)
{
	struct sharestruct *fs;
	int local_id;

	context;

	if(len < 22)
	{
		
		NOTICE("Not enough data in ft_fillstruct()!");
		return -1;
	}

	ft[id].fileid_len = AG_WORD(data);
	AG_POP_STR(ft[id].fileid, data, len);
	AG_POP_BYTE(ft[id].connecting, data, len);
	AG_POP_BYTE(ft[id].direction, data, len);

	AG_POP_STR(ft[id].ip, data, len);
	AG_POP_INT(ft[id].port, data, len);
	AG_POP_INT(local_id, data, len);
	AG_POP_STR(ft[id].file, data, len);
	AG_POP_INT(ft[id].filesize, data, len);

	ft[id].s = 0;
	ft[id].fd = NULL;
	ft[id].buf = NULL;

	/* AG_POP_* underflow? */
	if(len == -1)
	{
		if(ft[id].fileid)
			ag_free(ft[id].fileid);
		if(ft[id].ip)
			ag_free(ft[id].ip);
		if(ft[id].file)
			ag_free(ft[id].file);
		NOTICE("Not enough data in ft_fillstruct()!");
		return -1;
	}

	if(ft[id].direction == AG_FT_RECV)
	{
		ft[id].fdset = AG_FD_WRITE;
	}
	else
	{
		/* this always seems to be empty for uploads */
		/*if(ft[id].fileid)
			ag_free(ft[id].fileid);*/

		ft[id].fdset = AG_FD_READ;
		fs = share_find_by_id(local_id);
		if(!fs)
		{
			/* we dont have this file! */
			if(ft[id].ip)
				ag_free(ft[id].ip);
			if(ft[id].file)
				ag_free(ft[id].file);

			NOTICE("Server wants us to send a file we dont have!");
			/* TODO: somehow tell the server that we don't have this file? */

			ft_send_state(AG_FT_STATE_ERROR, ft[id].fileid, ft[id].fileid_len);
			return -1;
		}

		ft[id].ufile = ag_malloc(strlen(fs->path) + strlen(fs->file) + 2);
		sprintf(ft[id].ufile, "%s/%s", fs->path, fs->file);
	}

	NOTICE("Filetransfer: %s '%s' (%d bytes) %s %s:%d", ft[id].direction?"receive":"send",
			ft[id].file, ft[id].filesize, ft[id].direction?"from":"to",
			ft[id].ip, ft[id].port);

	return 0;
}


/* informs the server of the current state of a transfer */
int ft_send_state(int state, const unsigned char *fileid, int fileid_len)
{
	unsigned char *pkt,*ptr;
	int r;

	context;

	pkt = ptr = ag_malloc(2+fileid_len+4);
	if(!ptr) return -1;

	AG_PUSH_BIN(pkt, fileid, fileid_len);
	AG_PUSH_INT(pkt, state);
	r = ag_srv_msg(AG_MSG_FILETRANSFERSTATE, ptr, 2+fileid_len+4, QPRIO_HIGH);
	if(r == -1 && errno != EINPROGRESS)
		NOTICE("Unable to send filetransfer state to server: %s", ERRMSG());
	else
		r = 0;

	ag_free(ptr);

	return r;
}


/* sends a receive file request (REQ) */
int ft_send_request(int id, int resume_pos)
{
	unsigned char *pkt,*ptr;
	int r;

	context;

	pkt = ptr = ag_malloc(8+20+2+ft[id].fileid_len);
	if(!ptr) return -1;

	AG_PUSH_BYTE(pkt, 0x45);							/* header */
	AG_PUSH_BYTE(pkt, 0x5F);
	AG_PUSH_BYTE(pkt, 0xD0);
	AG_PUSH_BYTE(pkt, 0xD5);
	AG_PUSH_INT(pkt, 4+20+2+ft[id].fileid_len);			/* packet size */
	AG_PUSH_INT(pkt, 0);								/* ID */
	AG_PUSH_BIN(pkt, ft[id].fileid, ft[id].fileid_len);	/* file id */
	AG_PUSH_INT(pkt, ft[id].filesize);					/* file size */
	AG_PUSH_INT(pkt, resume_pos);						/* resume pos */

	r = send(ft[id].s, ptr, 8+20+2+ft[id].fileid_len, 0);

	if(r == -1 && errno != EINPROGRESS)
		NOTICE("Unable to send filetransfer request: %s", ERRMSG());
	else
		r = 0;

	ag_free(ptr);

	return r;
}


/* sends a send file response (ACK) */
int ft_send_response(int id)
{
	unsigned char *pkt,*ptr;
	int r,len;

	context;

	len = 4+4+4+2+2;
	pkt = ptr = ag_malloc(len);
	if(!ptr) return -1;

	AG_PUSH_BYTE(pkt, 0x45);	/* header */
	AG_PUSH_BYTE(pkt, 0x5F);
	AG_PUSH_BYTE(pkt, 0xD0);
	AG_PUSH_BYTE(pkt, 0xD5);
	AG_PUSH_INT(pkt, len-4);	/* packet size  */
	AG_PUSH_INT(pkt, 0);		/* ID  */
	AG_PUSH_STR(pkt, "OK");		/* "OK" */

	r = send(ft[id].s, ptr, len, 0);
	if(r == -1)
		NOTICE("Unable to send filetransfer response: %s", ERRMSG());
	else
		r = 0;

	ag_free(ptr);

	return r;
}


/*
	receives a receive file request (REQ)
	returns..
	-1	on fatal errors
	0	when there is not enough data
	>0	on success (returns the packet length)
*/
int ft_recv_request(int id, int len, unsigned char *data)
{
	unsigned char *fileid;
	int fileid_len,resume_pos,plen,filesize;

	context;

	/* we need at least 22 bytes. */
	if(len < 22)
		return 0;

	if(!AG_VALID_HDR(data) || !AG_VALID_PHDR(data))
	{
		NOTICE("INVALID PACKET HEADER! (%d, %d, %d, %d)",
				(int)data[0], (int)data[1], (int)data[2], (int)data[3]);
		return -1;
	}
	data += 4;	/* header */
	len -= 4;
	context;

	plen = AG_INT(data);
	if(plen > len)
	{
		DEBUG("need more data - %d - %d", plen, len);
		return 0;	/* we don't have the whole packet yet */
	}
	data += 8;	/* packet length + message count.. */
	len -= 8;
	context;

	fileid_len = AG_WORD(data);
	AG_POP_STR(fileid, data, len);
	if(!fileid)
		return -1;
	context;

	if(fileid_len != ft[id].fileid_len || memcmp(ft[id].fileid, fileid, fileid_len))
	{
		NOTICE("Fatal fileid mismatch. Aborting transfer of '%s'.", ft[id].file);
		return -1;
	}
	ag_free(fileid);

	AG_POP_INT(filesize, data, len);
	AG_POP_INT(resume_pos, data, len);

	context;
	if(resume_pos < 0)
	{
		NOTICE("Remote client sent negative resume position. Aborting transfer of '%s'.", ft[id].file);
		return -1;
	}

	if(resume_pos >= ft[id].filesize)
	{
		NOTICE("Remote client wants to resume beyond end of file. Aborting transfer of '%s'.", ft[id].file);
		return -1;
	}

	if(resume_pos > 0)
	{
		NOTICE("Resuming transfer of '%s' at %d bytes.", ft[id].file, resume_pos);
		fseek(ft[id].fd, resume_pos, SEEK_SET);
		ft[id].lastspd_pos = resume_pos;
		/*ft[id].position = resume_pos;*/
	}

	if(ft[id].filesize != filesize)
	{
		NOTICE("Filesize mismatch: File is %d bytes long, "
				"remote client says it's %d bytes.", ft[id].filesize, filesize);
		if(resume_pos)
		{
			NOTICE("Invalid file size and non-zero resume position. Aborting transfer of '%s'.", ft[id].file);
			return -1;
		}
	}

	if(len < 0)
		/* not enough data ... */
		return -1;

	context;
	return plen;
}


/*
	receives a send file response (ACK)
	returns..
	-1	on fatal errors
	0	when there is not enough data
	>0	on success (returns the packet length)
*/
int ft_recv_response(int id, int len, unsigned char *data)
{
	unsigned char *okstr;
	int plen;

	context;

	/* we need at least 17 bytes. */
	if(len < 17)
		return 0;

	if(!AG_VALID_HDR(data) || !AG_VALID_PHDR(data))
	{
		NOTICE("INVALID PACKET HEADER! (%d, %d, %d, %d)",
				(int)data[0], (int)data[1], (int)data[2], (int)data[3]);
		return -1;
	}
	data += 4;
	len -= 4;

	if(AG_INT(data) > len)
	{
		DEBUG("need more data - %d - %d", AG_INT(data), len);
		return 0;	/* we don't have the whole packet yet */
	}
	AG_POP_INT(plen, data, len);
	if(plen <= 0 || plen >= 1024) {
		NOTICE("Strange packet size. Aborting transfer of '%s'.", ft[id].file);
		return -1;
	}

	/* ID */
	data += 4;
	len -= 4;

	AG_POP_STR(okstr, data, len);
	if(!okstr || strcmp("OK", okstr))
	{
		if(okstr)
			ag_free(okstr);
		NOTICE("Invalid 'OK' string. Aborting transfer of '%s'.", ft[id].file);
		return -1;
	}
	ag_free(okstr);

	if(len < 0)
		/* not enough data ... */
		return -1;

	return plen;
}
