
#define MALLOC_MAGIC1	0xDE
#define MALLOC_MAGIC2	0xAD
#define MALLOC_MAGIC3	0xBE
#define MALLOC_MAGIC4	0xEF

struct malloc_list
{
	struct malloc_list *prev;
	struct malloc_list *next;
	void *ptr;
};


struct malloc_list *mem_list=NULL;
struct malloc_list *mem_list_cur=NULL;


struct malloc_list *mem_list_new()
{
	if(mem_list == NULL)
	{
		mem_list = mem_list_cur = malloc(sizeof(struct malloc_list));
		mem_list->prev = NULL;
		mem_list_cur->next = NULL;
	}
	else
	{
		mem_list_cur->next = malloc(sizeof(struct malloc_list));
		mem_list_cur->next->prev = mem_list_cur;
		mem_list_cur = mem_list_cur->next;
		mem_list_cur->next = NULL;
	}

	return mem_list_cur;
}


void mem_list_del(struct malloc_list *m)
{
	if(!m) return;

	if(m->next)
		m->next->prev = m->prev;
	if(m->prev)
		m->prev->next = m->next;

	if(!m->next)
		mem_list_cur = m->prev;

	if(!m->prev)
		mem_list = m->next;

	free(m);
}


struct malloc_list *mem_list_find(void *ptr)
{
	struct malloc_list *m;

	m = mem_list_cur;
	while(m)
	{
		if(m->ptr == ptr)
			return m;
		if(m == mem_list)
			break;
		m = m->prev;
	}

	return NULL;
}


void malloc_dump()
{
	struct malloc_list *m = mem_list;
	unsigned char *ptr;
	int size;

	printf("Memory still allocated:\n");
	while(m)
	{
		ptr = m->ptr;

		size = ptr[0] + ptr[1]*256;

		printf("%d bytes malloc()ed at %s, not freed.\n",
				size, &ptr[size+10+128]);

		m = m->next;
	}
}


void malloc_free_all()
{
	struct malloc_list *m,*n;
	void *ptr;
	int c=0;

	m = mem_list;
	while(m)
	{
		n = m->next;
		if(m->ptr)
		{
			c++;
			ptr = m->ptr+6;
			ag_free(ptr);
			mem_list_del(m);
		}
		m = n;
	}
	printf("Freed %d pointers.\n", c);

	return;
}


void *_ag_malloc(int size, unsigned char *file, int line)
{
	static unsigned char *ptr;

	if(size > 32768)
		return NULL;

	ptr = malloc(2+4+size+4+32+128);
	if(!ptr)
	{
		printf("malloc() returned NULL ptr! (%s:%d)\n", file, line);
	}
	else
	{
		/*if(mem_list_find(ptr))
			printf("malloc() returned existing ptr %d @ %s:%d\n", (int)ptr, file, line);*/
		ptr[0] = size%256;
		ptr[1] = size/256;
		ptr[2] = MALLOC_MAGIC1;
		ptr[3] = MALLOC_MAGIC2;
		ptr[4] = MALLOC_MAGIC3;
		ptr[5] = MALLOC_MAGIC4;
		ptr[size+6] = MALLOC_MAGIC4;
		ptr[size+7] = MALLOC_MAGIC3;
		ptr[size+8] = MALLOC_MAGIC2;
		ptr[size+9] = MALLOC_MAGIC1;

		snprintf(&ptr[size+10+128], 32, "%s:%d (%d)", file, line, size);

		mem_list_new()->ptr = ptr;
	}

	return (void *)(ptr+6);
}

void _ag_free(void *_ptr, unsigned char *file, int line)
{
	unsigned char *ptr = _ptr;
	struct malloc_list *m;
	int size,ov=0;

	ptr -= 6;

	if(!ptr)
	{
		printf("Tried to free() NULL ptr @ %s:%d\n", file, line);
		return;
	}

	m = mem_list_find(ptr);
	if(!m)
	{
		printf("Tried to free() unallocated ptr %d @ %s:%d\n", (int)ptr, file, line);
		return;
	}

	size = ptr[0] + ptr[1]*256;
	if(ptr[2] != MALLOC_MAGIC1) ov |= 1;
	if(ptr[3] != MALLOC_MAGIC2) ov |= 1;
	if(ptr[4] != MALLOC_MAGIC3) ov |= 1;
	if(ptr[5] != MALLOC_MAGIC4) ov |= 1;
	if(ptr[size+6] != MALLOC_MAGIC4) ov |= 1;
	if(ptr[size+7] != MALLOC_MAGIC3) ov |= 1;
	if(ptr[size+8] != MALLOC_MAGIC2) ov |= 1;
	if(ptr[size+9] != MALLOC_MAGIC1) ov |= 1;

	if(ov)
	{
		printf("free() %d bytes of memory @ %s:%d (allocated at %s) - %s\n",
			size, file, line, &ptr[size+10+128], ov?"Memory was overwritten!!!":"Memory OK.");
		sleep(1);
		printf("magic should be [%d,%d,%d,%d] and [%d,%d,%d,%d]\n",
				MALLOC_MAGIC1, MALLOC_MAGIC2, MALLOC_MAGIC3, MALLOC_MAGIC4,
				MALLOC_MAGIC4, MALLOC_MAGIC3, MALLOC_MAGIC2, MALLOC_MAGIC1);
		printf("       magic is [%d,%d,%d,%d] and [%d,%d,%d,%d]\n",
				ptr[2], ptr[2], ptr[3], ptr[4],
				ptr[size+6], ptr[size+7], ptr[size+8], ptr[size+9]);
	}

	free(ptr);

	mem_list_del(m);

	return;
}

unsigned char *_ag_strdup(const unsigned char *s, unsigned char *file, int line)
{
	unsigned char *ptr;
	int size;

	if(!s)
	{
		printf("Tried to strdup() NULL ptr! (%s:%d)\n", file, line);
		return NULL;
	}

	size = strlen(s)+1;
	ptr = malloc(strlen(s)+1+2+8+128+32);
	if(!ptr)
	{
		printf("strdup() returned NULL ptr! (%s:%d)\n", file, line);
		return NULL;
	}
	else
	{
		if(mem_list_find(ptr))
			printf("malloc() returned existing ptr %d @ %s:%d\n", (int)ptr, file, line);

		ptr[0] = size%256;
		ptr[1] = size/256;
		ptr[2] = MALLOC_MAGIC1;
		ptr[3] = MALLOC_MAGIC2;
		ptr[4] = MALLOC_MAGIC3;
		ptr[5] = MALLOC_MAGIC4;
		strcpy(&ptr[6], s);
		ptr[size+6] = MALLOC_MAGIC4;
		ptr[size+7] = MALLOC_MAGIC3;
		ptr[size+8] = MALLOC_MAGIC2;
		ptr[size+9] = MALLOC_MAGIC1;

		snprintf(&ptr[size+10+128], 32, "*%s:%d (%d)", file, line, size);

		mem_list_new()->ptr = ptr;
	}

	return ptr+6;
}


