#include "config.h"
#include "macros.h"
#include "network.h"
#include "ag.h"
#include "queue.h"

struct queue_struct *queue[2]={NULL,NULL};
struct queue_struct *queue_last[2]={NULL,NULL};


void queue_add(unsigned char *packet, int len, int sent, int prio)
{
	struct queue_struct *q;

	context;

	/*DEBUG("queueing packet len %d prio %d (%d sent)", len, prio, sent);*/

	q = queue_last[prio];
	if(!q)
	{
		q = queue[prio] = ag_malloc(sizeof(struct queue_struct));
	}
	else
	{
		q->next = ag_malloc(sizeof(struct queue_struct));
		q = q->next;
	}
	q->next = NULL;
	q->len = len;
	q->packet = packet;
	q->sent = sent;

	queue_last[prio] = q;
}


int queue_poll()
{
	struct queue_struct *q, *tmp;
	int r, prio=QPRIO_HIGH;

	context;

	q = queue[prio];
	while(1)
	{
		if(q)
		{
			/*DEBUG("sending queued packet len %d prio %d (%d sent)", q->len, prio, q->sent);*/
			r = send(srv_s, q->packet + q->sent, q->len - q->sent, 0);
			if(r > 0 && r < (q->len - q->sent))
			{
				q->sent += r;
				DEBUG("partial queued packet sent len %d (%d left)", q->len, q->len - q->sent);
				return 1;
			}
			else if(r <= 0)
			{
				if(errno == EAGAIN || errno == EWOULDBLOCK || errno == ENOBUFS)
				{
					DEBUG("couldn't send queued packet");
					return 1;
				}
				else
				{
					NOTICE("Error while sending queued packet to server: %s", ERRMSG());
					queue_empty();
					return 0;
				}
			}
			if(q->sent == q->len)
				DEBUG("partial queued packet successfully sent len %d", q->len);
			tmp = q->next;
			ag_free(q->packet);
			ag_free(q);
			q = queue[prio] = tmp;
			if(q == NULL) queue_last[prio] = NULL;
		}
		else if(prio == QPRIO_HIGH)
		{
			prio = QPRIO_LOW;
			q = queue[prio];
		}
		else
		{
			return 1;
		}
	}
}


void _queue_empty(struct queue_struct *q)
{
	struct queue_struct *tmp;

	context;

	while(q)
	{
		tmp = q->next;
		ag_free(q->packet);
		ag_free(q);
		q = tmp;
	}
}


void queue_empty()
{
	context;

	_queue_empty(queue[QPRIO_LOW]);
	_queue_empty(queue[QPRIO_HIGH]);

	queue[QPRIO_LOW] = NULL;
	queue[QPRIO_HIGH] = NULL;
	queue_last[QPRIO_LOW] = NULL;
	queue_last[QPRIO_HIGH] = NULL;
}


int queue_is_empty()
{
	if(queue[QPRIO_LOW] || queue[QPRIO_HIGH])
		return 0;

	return 1;
}

